class NamespaceDict(dict):
    def __init__(self, *args, **kwargs):
        super(NamespaceDict, self).__init__(*args, **kwargs)
        for arg in args:
            if isinstance(arg, dict):
                for k, v in arg.items():
                    self[k] = v
        if kwargs:
            for k, v in kwargs.items():
                self[k] = v

    def __getattr__(self, attr):
        return self.get(attr)

    def __getitem__(self, item):
        idx_dot = item.find('.')
        if idx_dot >= 0:
            curr_attr = item[:idx_dot]
            nested_attr = item[idx_dot + 1:]
            nested_map = self.get(curr_attr)
            return nested_map.get(nested_attr)
        else:
            return self.get(item)

    def __setitem__(self, key, value):
        if isinstance(value, dict):
            value = NamespaceDict(value)
        idx_dot = key.find('.')
        if idx_dot >= 0:
            attr = key[:idx_dot]
            nested_key = key[idx_dot+1:]
            nested_map = self.get(attr)
            if nested_map is None:
                nested_map = NamespaceDict()
            nested_map.__setitem__(nested_key, value)
            super(NamespaceDict, self).__setitem__(attr, nested_map)
            self.__dict__.update({attr: nested_map})
        else:
            super(NamespaceDict, self).__setitem__(key, value)
            self.__dict__.update({key: value})

    def __delattr__(self, item):
        self.__delitem__(item)

    def __delitem__(self, key):
        super(NamespaceDict, self).__delitem__(key)
        del self.__dict__[key]




import pytest

from argumento.namespace_dict import NamespaceDict


def test_set_get_simple():
    ns_dict = NamespaceDict()
    ns_dict['item'] = 1
    assert ns_dict['item'] == 1
    assert ns_dict.item == 1


def test_init_from_dict():
    ns_dict = NamespaceDict({'item1': 1,
                             'item2': 2,
                             'item3': {
                                 'subitem31': {'subitem311': 311, 'subitem312': 312},
                                 'subitem32': {'subitem321': 321, 'subitem322': 322}}})
    assert ns_dict['item1'] == 1
    assert ns_dict['item2'] == 2
    assert ns_dict.item1 == 1
    assert ns_dict.item2 == 2

    assert ns_dict['item3']['subitem31']['subitem311'] == 311
    assert ns_dict['item3']['subitem31']['subitem312'] == 312
    assert ns_dict['item3']['subitem32']['subitem321'] == 321
    assert ns_dict['item3']['subitem32']['subitem322'] == 322
    #
    assert ns_dict.item3['subitem31']['subitem311'] == 311
    assert ns_dict.item3['subitem31']['subitem312'] == 312
    assert ns_dict.item3['subitem32']['subitem321'] == 321
    assert ns_dict.item3['subitem32']['subitem322'] == 322
    #
    assert ns_dict.item3.subitem31['subitem311'] == 311
    assert ns_dict.item3.subitem31['subitem312'] == 312
    assert ns_dict.item3.subitem32['subitem321'] == 321
    assert ns_dict.item3.subitem32['subitem322'] == 322
    #
    assert ns_dict.item3.subitem31.subitem311 == 311
    assert ns_dict.item3.subitem31.subitem312 == 312
    assert ns_dict.item3.subitem32.subitem321 == 321
    assert ns_dict.item3.subitem32.subitem322 == 322

def test_set_recursive():
    ns_dict = NamespaceDict()
    ns_dict['item1.subitem1'] = 11
    assert ns_dict.item1.subitem1 == 11
    assert ns_dict['item1.subitem1'] == 11

    ns_dict['item1']['subitem2'] = 12
    assert ns_dict['item1']['subitem2'] == 12
    assert ns_dict['item1.subitem2'] == 12
    assert ns_dict.item1.subitem2 == 12

    with pytest.raises(TypeError):
        ns_dict['item2']['subitem2'] = 22

    ns_dict['item2'] = {'subitem2': 22}
    assert ns_dict['item2']['subitem2'] == 22
    assert ns_dict['item2.subitem2'] == 22
    assert ns_dict.item2.subitem2 == 22

    ns_dict['item3'] = {'subitem31': {'subitem311': 311, 'subitem312': 312},
                        'subitem32': {'subitem321': 321, 'subitem322': 322}}
    assert ns_dict['item3']['subitem31']['subitem311'] == 311
    assert ns_dict['item3']['subitem31']['subitem312'] == 312
    assert ns_dict['item3']['subitem32']['subitem321'] == 321
    assert ns_dict['item3']['subitem32']['subitem322'] == 322
    #
    assert ns_dict.item3['subitem31']['subitem311'] == 311
    assert ns_dict.item3['subitem31']['subitem312'] == 312
    assert ns_dict.item3['subitem32']['subitem321'] == 321
    assert ns_dict.item3['subitem32']['subitem322'] == 322
    #
    assert ns_dict.item3.subitem31['subitem311'] == 311
    assert ns_dict.item3.subitem31['subitem312'] == 312
    assert ns_dict.item3.subitem32['subitem321'] == 321
    assert ns_dict.item3.subitem32['subitem322'] == 322
    #
    assert ns_dict.item3.subitem31.subitem311 == 311
    assert ns_dict.item3.subitem31.subitem312 == 312
    assert ns_dict.item3.subitem32.subitem321 == 321
    assert ns_dict.item3.subitem32.subitem322 == 322


import sys
import os
import pytest
from unittest import mock
from argumento import create_parser


EXTENSIONS = ['toml', 'yaml', 'yml', 'json']


def locate_data_file(filename):
    return os.path.join('tests', 'data', filename)


@pytest.mark.parametrize('ext', EXTENSIONS)
def test_flat(ext):
    cfg_filename = locate_data_file(f'flat.{ext}')
    args = create_parser(cfg_filename).parse()

    assert args.login == "my_login"
    assert args.max_retries == 5
    assert args.ports == [8000, 8001, 8002]
    assert args.ratio == 0.5


@pytest.mark.parametrize('ext', EXTENSIONS)
def test_flat_cmd_override(ext):
    cfg_filename = locate_data_file(f'flat.{ext}')
    cmd_args = ['', '--login', 'login_from_cmd', '--max_retries', '3',
                '--nice_bool_false', 'False', '--nice_bool_true', 'true']
    with mock.patch.object(sys, 'argv', cmd_args):
        args = create_parser(cfg_filename).parse()

        # these args were overridden in command-line
        assert args.login == "login_from_cmd"
        assert args.max_retries == 3
        assert args.nice_bool_false == False
        assert args.nice_bool_true == True

        # these args were read from config file
        assert args.ports == [8000, 8001, 8002]
        assert args.ratio == 0.5


@pytest.mark.parametrize('ext', EXTENSIONS)
def test_redundant_cmd_args(ext):
    """
    should not FAIL if args in command line are not specified in config file
    :return:
    :rtype:
    """
    cfg_filename = locate_data_file(f'flat.{ext}')
    cmd_args = ['', '--login', 'login_from_cmd', '--max_retries', '3', '--redundant-arg', '123']
    with mock.patch.object(sys, 'argv', cmd_args):
        args = create_parser(cfg_filename).parse()


@pytest.mark.parametrize('ext', EXTENSIONS)
def test_flat_cmd_only(ext):
    cfg_filename = locate_data_file(f'flat_cmd_only.{ext}')
    cmd_args = ['', '--login', 'login_from_cmd', '--max_retries', '3', '--ratio', '0.2', '--ports', '5000,5001',
                '--fractions', '0.1, 0.2']
    with mock.patch.object(sys, 'argv', cmd_args):

        args = create_parser(cfg_filename).parse()

        assert args.login == "login_from_cmd"
        assert args.max_retries == 3
        assert args.ratio == 0.2
        assert args.ports == [5000, 5001]
        assert args.fractions == [0.1, 0.2]


@pytest.mark.parametrize('ext', EXTENSIONS)
def test_hierarch(ext):
    cfg_filename = locate_data_file(f'hierarch.{ext}')
    args = create_parser(cfg_filename).parse()

    assert args.database == {"ports": [ 8000, 8001, 8002 ], "connection_max": 5000, "enabled": True}
    assert args['database'] == args.database
    assert args.database.ports == [8000, 8001, 8002]
    assert args.database.ports == args.database['ports']
    assert args.database.ports == args['database'].ports
    assert args.database.ports == args['database.ports']
    assert args.database.connection_max == 5000
    assert args.database.connection_max == args.database['connection_max']
    assert args.database.connection_max == args['database'].connection_max
    assert args.database.connection_max == args['database.connection_max']

    assert args.database.enabled == True

    assert args.servers.alpha == {'ip': '10.0.0.1', 'dc': 'abcd'}
    assert args.servers.alpha.ip == '10.0.0.1'
    assert args.servers.alpha.dc == 'abcd'

    assert args.servers.beta == {'ip': '10.0.0.2', 'dc': 'efgh'}
    assert args.servers.beta.ip == '10.0.0.2'
    assert args.servers.beta.dc == 'efgh'

# TODO: test hierarchical config - overwrite params in cmd


def setup_env():
    os.environ['PORT_1'] = '8000'
    os.environ['PORT_2'] = '8001'
    os.environ['PORT_3'] = '8002'
    os.environ['ENABLED'] = 'True'
    # os.environ['DISABLED'] = 'False'  # not set in ENV - should be None
    os.environ['CONNECTION_MAX'] = '5000'

    os.environ['IP_ALPHA'] = '10.0.0.1'
    os.environ['DC_ALPHA'] = 'ab cd'
    # os.environ['IP_BETA'] = '10.0.0.1:1337'  # not set in ENV
    os.environ['DC_BETA'] = 'xyz'


@pytest.mark.parametrize('ext', EXTENSIONS)
def test_env(ext):
    setup_env()
    cfg_filename = locate_data_file(f'env.{ext}')
    args = create_parser(cfg_filename).parse()

    assert args.database == {"ports": ['8000', '8001', '8002'],
                             "ports_other": ['8000', '8001', '8002'],
                             "enabled": 'True', "disabled": '',
                             "connection_max": '5000'}
    assert args['database'] == args.database
    assert args.database.ports == ['8000', '8001', '8002']
    assert args.database.ports == args.database['ports']
    assert args.database.ports == args['database'].ports
    assert args.database.ports == args['database.ports']
    assert args.database.connection_max == '5000'
    assert args.database.connection_max == args.database['connection_max']
    assert args.database.connection_max == args['database'].connection_max
    assert args.database.connection_max == args['database.connection_max']

    assert args.database.enabled == 'True'
    assert args.database.enabled == 'True'
    assert args.servers.alpha == {'ip': '10.0.0.1', 'dc': 'path/ab cd/other_path/qwe'}
    assert args.servers.alpha.ip == '10.0.0.1'
    assert args.servers.alpha.dc == "path/ab cd/other_path/qwe"

    assert args.servers.beta == {'ip': '10.0.0.2:1337', 'dc': 'path/xyz/other_path'}
    assert args.servers.beta.ip == '10.0.0.2:1337'
    assert args.servers.beta.dc == "path/xyz/other_path"





import argparse
import json
import os.path
import os
import re
from abc import ABC, abstractmethod
from collections import abc
from typing import Union, List, Type
import warnings

import toml
import yaml
import builtins

from argumento.namespace_dict import NamespaceDict


def str2bool(v):
    if isinstance(v, bool):
        return v
    if v.lower() in ('yes', 'true', 't', 'y', '1'):
        return True
    elif v.lower() in ('no', 'false', 'f', 'n', '0'):
        return False
    else:
        raise argparse.ArgumentTypeError('Boolean value expected.')


class EnvResolver:
    def __init__(self):
        self.default_value = None
        self.pattern = re.compile(r'\$\{(\w+)(?:\|([^}]*))?\}')  # pattern: '${NAME|default}'

    def _nested_dict_iter(self, nested: abc.Mapping):
        """
        Recursively iterates over nested (optionally) dictionary,
          traversing list-like and dict-like structures,
          yields on leaf nodes (to be processed).
        """
        for key, value in nested.items():
            if isinstance(value, abc.Mapping):
                yield from self._nested_dict_iter(value)
            elif isinstance(value, list):
                for i, item in enumerate(value):
                    if isinstance(item, abc.Mapping):
                        yield from self._nested_dict_iter(item)
                    else:
                        yield value, i, item
            else:
                yield nested, key, value

    def _parse_string(self, value: str):
        """Pattern-matches string with self.pattern, substitutes match groups with os.environ/default values"""
        def replacer(match):
            var_name = match.group(1)
            default_value = match.group(2) if match.group(2) is not None else self.default_value
            return os.environ.get(var_name, default_value)
        return self.pattern.sub(replacer, value)

    def resolve_envs(self, config: dict) -> dict:
        """Entry point for config (dict) variables recursive env. search and resolve"""
        for container, key, value in self._nested_dict_iter(config):
            if isinstance(value, str):
                container[key] = self._parse_string(value)
        return config


class ParserBase(ABC):
    def __init__(self, config_file: str):
        self._config_file = config_file
        self.env_resolver = EnvResolver()

    @abstractmethod
    def _read_config(self):
        pass

    def parse(self) -> NamespaceDict:
        cfg_file_dict = self._read_config()
        cfg_file_dict = self.env_resolver.resolve_envs(cfg_file_dict)
        cmd_arg_parser = self._cmd_args_from_cfg_keys(cfg_file_dict)
        args, _ = cmd_arg_parser.parse_known_args()
        return NamespaceDict(vars(args))

    @classmethod
    def _cmd_args_from_cfg_keys(cls,
                                cfg_file_dict: dict,
                                parent_cfg_name: str = None,
                                cmd_parser: argparse.ArgumentParser = None
                                ) -> argparse.ArgumentParser:
        if cmd_parser is None:
            cmd_parser = argparse.ArgumentParser()
        for k, v in cfg_file_dict.items():
            full_key_name = k if parent_cfg_name is None else f'{parent_cfg_name}.{k}'
            if isinstance(v, dict):
                # recursion for nested nodes in config
                cls._cmd_args_from_cfg_keys(v, full_key_name, cmd_parser)
            elif isinstance(v, str):
                v_no_space = v.replace(' ', '')
                if v_no_space.startswith('?:'):
                    # required param, not set in the config file, needs top be set in command line
                    type_name = v_no_space[2:]
                    # only list[int], list[float] supported
                    # TODO: list[str]
                    if type_name.startswith('list[') and type_name.endswith(']'):
                        item_type_name = type_name[5:-1]
                        assert item_type_name in ['int', 'float']
                        item_type = getattr(builtins, item_type_name)
                        cmd_parser.add_argument(f'--{full_key_name}',
                                                type=lambda s: [item_type(item) for item in s.split(',')])
                    else:
                        type_ = getattr(builtins, type_name)
                        cmd_parser.add_argument(f'--{full_key_name}', required=True, type=type_)
                else:
                    cmd_parser.add_argument(f'--{full_key_name}', type=str, default=v)
            elif isinstance(v, bool):
                cmd_parser.add_argument(f'--{full_key_name}', type=str2bool, default=v)
            else:
                cmd_parser.add_argument(f'--{full_key_name}', type=type(v), default=v)
        return cmd_parser


class ParserToml(ParserBase):
    def __init__(self, config_file: str):
        super().__init__(config_file)

    def _read_config(self) -> dict:
        with open(self._config_file, 'r', encoding='utf8') as f:
            return toml.load(f)


class ParserYaml(ParserBase):
    def __init__(self, config_file: str):
        super().__init__(config_file)

    def _read_config(self) -> dict:
        with open(self._config_file, 'r', encoding='utf8') as f:
            return yaml.safe_load(f)


class ParserJson(ParserBase):
    def __init__(self, config_file: str):
        super().__init__(config_file)

    def _read_config(self) -> dict:
        with open(self._config_file, 'r', encoding='utf8') as f:
            data = f.read().encode("utf-8")
            return json.loads(data)


class ParserFactory:
    def __init__(self):
        self._parsers = {}

    def _register_single_format(self, fmt: str, parser: Type[ParserBase]):
        if fmt in self._parsers:
            warnings.warn(f'Format {fmt} is already bound to {self._parsers[fmt]}. The value will be overwritten!')
        self._parsers[fmt] = parser

    def register_format(self, fmt: Union[str, List[str]], parser_type: Type[ParserBase]):
        if isinstance(fmt, str):
            self._register_single_format(fmt, parser_type)
        elif isinstance(fmt, list) and all(isinstance(x, str) for x in fmt):
            for fmt_item in fmt:
                self._register_single_format(fmt_item, parser_type)
        else:
            raise ValueError(f'Invalid format: {fmt}')

    def get_parser(self, filename: str) -> ParserBase:
        ext = os.path.splitext(filename)[1]
        if ext.startswith('.'):
            ext = ext[1:]
        parser_type = self._parsers.get(ext)
        if not parser_type:
            raise ValueError(f'Could not infer parser from file type. Unknown file format: "{ext}". '
                             f'Supported formats are: {list(self._parsers.keys())}.')
        return parser_type(filename)


factory = ParserFactory()
factory.register_format(['yml', 'yaml'], ParserYaml)
factory.register_format('toml', ParserToml)
factory.register_format('json', ParserJson)

create_parser = factory.get_parser




import pytest

import argumento


def test_unknown_format():
    # .ini not supported
    with pytest.raises(ValueError) as exc_info:
        argumento.create_parser('my_file.ini').parse()

    assert exc_info.value.args[0] == "Could not infer parser from file type. Unknown file format: \"ini\". " \
                                     "Supported formats are: ['yml', 'yaml', 'toml', 'json']."


def test_reqister_custom_parser(tmp_path):

    class ParserCustomTest(argumento.parsers.ParserBase):
        """

        custom cfg format:

        key1: value1
        key2: value2
        ...

        """
        def __init__(self, config_file: str):
            super().__init__(config_file)

        def _read_config(self) -> dict:
            with open(self._config_file, "r") as f:
                lines = [l for l in f if ':' in l]
                return dict([x.strip() for x in l.split(':')] for l in lines)

    # assert argumento.create_parser is argumento.factory.get_parser

    factory = argumento.parsers.ParserFactory()
    factory.register_format('custom', ParserCustomTest)

    cfg_filename = "cfg.custom"

    d = tmp_path / "sub"
    d.mkdir()
    p = d / cfg_filename
    content = """
    key1: value1
    key2: value2
    """
    p.write_text(content)

    cfg = factory.get_parser(p).parse()
    assert vars(cfg) == {'key1': 'value1', 'key2': 'value2'}
